#include <wdm.h>

#include "Hypervisor.h"

namespace
{
    UNICODE_STRING DeviceName = RTL_CONSTANT_STRING(L"\\Device\\Hypervisor");
    UNICODE_STRING DeviceLink = RTL_CONSTANT_STRING(L"\\??\\Hypervisor");
    PDEVICE_OBJECT DeviceInstance = NULL;
}

EXTERN_C_START

DRIVER_INITIALIZE DriverEntry;

_Function_class_(DRIVER_DISPATCH)
_Dispatch_type_(IRP_MJ_DEVICE_CONTROL)
_IRQL_requires_max_(DISPATCH_LEVEL)
_IRQL_requires_same_
static NTSTATUS NTAPI DriverControl(
    _In_ PDEVICE_OBJECT DeviceObject,
    _In_ PIRP Irp
);

_Function_class_(DRIVER_DISPATCH)
_Dispatch_type_(IRP_MJ_CREATE)
_Dispatch_type_(IRP_MJ_CLOSE)
_Dispatch_type_(IRP_MJ_CLEANUP)
_IRQL_requires_max_(DISPATCH_LEVEL)
_IRQL_requires_same_
static NTSTATUS NTAPI DriverStub(
    _In_ PDEVICE_OBJECT DeviceObject,
    _In_ PIRP Irp
);

_Function_class_(DRIVER_UNLOAD)
_IRQL_requires_(PASSIVE_LEVEL)
_IRQL_requires_same_
static NTSTATUS NTAPI DriverUnload(
    _In_ PDRIVER_OBJECT DriverObject
);

EXTERN_C_END

extern "C" NTSTATUS NTAPI DriverEntry(
    _In_ PDRIVER_OBJECT DriverObject,
    _In_ PUNICODE_STRING RegistryPath
) {
    UNREFERENCED_PARAMETER(RegistryPath);

    DriverObject->DriverUnload = reinterpret_cast<PDRIVER_UNLOAD>(DriverUnload);
    DriverObject->MajorFunction[IRP_MJ_CREATE] = DriverStub;
    DriverObject->MajorFunction[IRP_MJ_CLEANUP] = DriverStub;
    DriverObject->MajorFunction[IRP_MJ_CLOSE] = DriverStub;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = DriverControl;
    
    NTSTATUS Status = IoCreateDevice(DriverObject, 0, &DeviceName, FILE_DEVICE_UNKNOWN, 0, FALSE, &DeviceInstance);
    if (!NT_SUCCESS(Status))
    {
        DbgPrint("IoCreateDevice error: 0x%X!\r\n", Status);
        return Status;
    }
    
    Status = IoCreateSymbolicLink(&DeviceLink, &DeviceName);
    if (!NT_SUCCESS(Status))
    {
        DbgPrint("IoCreateSymbolicLink error: 0x%X!\r\n", Status);
        IoDeleteDevice(DeviceInstance);
        return Status;
    }
    
    // Let's virtualize it:
    bool VirtualizationStatus = Hypervisor::Virtualize();
    DbgPrint("Virtualization status: %u", static_cast<unsigned int>(VirtualizationStatus));

    return STATUS_SUCCESS;
}

_Function_class_(DRIVER_DISPATCH)
_Dispatch_type_(IRP_MJ_DEVICE_CONTROL)
_IRQL_requires_max_(DISPATCH_LEVEL)
_IRQL_requires_same_
static NTSTATUS NTAPI DriverControl(_In_ PDEVICE_OBJECT DeviceObject, _In_ PIRP Irp)
{
    UNREFERENCED_PARAMETER(DeviceObject);

    // ... Put your custom code here if you need it ...

    Irp->IoStatus.Status = STATUS_SUCCESS;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    
    return STATUS_SUCCESS;
}

_Function_class_(DRIVER_DISPATCH)
_Dispatch_type_(IRP_MJ_CREATE)
_Dispatch_type_(IRP_MJ_CLOSE)
_Dispatch_type_(IRP_MJ_CLEANUP)
_IRQL_requires_max_(DISPATCH_LEVEL)
_IRQL_requires_same_
static NTSTATUS NTAPI DriverStub(_In_ PDEVICE_OBJECT DeviceObject, _In_ PIRP Irp)
{
    UNREFERENCED_PARAMETER(DeviceObject);
    
    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;
    IoCompleteRequest(Irp, IO_NO_INCREMENT);
    
    return STATUS_SUCCESS;
}

_Function_class_(DRIVER_UNLOAD)
_IRQL_requires_(PASSIVE_LEVEL)
_IRQL_requires_same_
static NTSTATUS NTAPI DriverUnload(_In_ PDRIVER_OBJECT DriverObject)
{
    UNREFERENCED_PARAMETER(DriverObject);

    Hypervisor::Devirtualize();

    IoDeleteSymbolicLink(&DeviceLink);
    IoDeleteDevice(DeviceInstance);

    DbgPrint("Successfully unloaded!\r\n");
    return STATUS_SUCCESS;
}